# Copyright (C) 2017-2021 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from six.moves import input
import sys

from .ipaddresses import IPAddress

class Ask(object):

    def __init__(self):
        self.next_answer = None

    def get_input(self):
        if self.next_answer:
            ans = self.next_answer
            self.next_answer = None
            self.wr_out(ans + "\n")
            return ans
        else:
            return input()

    def wr_out(self, msg):
        sys.stdout.write(msg)

    def ask(self, text, options=None, default=None, raise_exc=False):
        """Returns: ([True|False], answer)"""

        msg = "{} ".format(text)
        if options:
            msg_options = []
            for opt in options:
                if opt == default:
                    msg_options.append(opt.upper())
                else:
                    msg_options.append(str(opt))
            msg += "["
            msg += "/".join(msg_options)
            msg += "] "
        else:
            if default:
                msg += "(default: {}) ".format(default)
        self.wr_out(msg)

        try:
            answer = self.get_input()
        except:
            if raise_exc:
                raise
            return False, None

        answer = answer.strip()
        if answer:
            if options and answer.lower() not in [_.lower() for _ in options]:
                print("Invalid choice: {} - must be one of {}.".format(
                    answer, ", ".join(options)))
                return False, None
            return True, answer
        else:
            if default:
                return True, default
            else:
                print("No answer given.")
                return False, None

    def ask_yes_no(self, text, default=None, raise_exc=False):
        return self.ask(text, ["yes", "no"], default, raise_exc)

    def ask_int(self, text, default=None, raise_exc=False):
        answer_given, v = self.ask(text, None, default, raise_exc)
        if not answer_given:
            return False, None
        if not v.isdigit():
            print("Invalid input: it must be an integer.")
            return False, None
        return True, int(v)

    def ask_ipv4_addr(self, text, default=None, raise_exc=False):
        answer_given, v = self.ask(text, None, default, raise_exc)
        if not answer_given:
            return False, None
        try:
            ip = IPAddress(v)
            if ip.version != 4:
                raise ValueError()
        except:
            print("Invalid input: must be a valid IPv4 address.")
            return False, None
        return True, ip.ip



# Copyright (C) 2017-2021 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from .docker import DockerInstance
from .instances import Route, BGPSpeakerInstance, InstanceNotRunning


class BIRDInstance(DockerInstance, BGPSpeakerInstance):
    """This class implements BIRD-specific methods.

    This class is derived from :class:`DockerInstance`, that implements
    some Docker-specific methods to start/stop the instance and to run
    commands on it.
    """

    MESSAGE_LOGGING_SUPPORT = True

    DOCKER_IMAGE = "pierky/bird:1.6.8"

    TAG = "bird16"

    def __init__(self, *args, **kwargs):
        DockerInstance.__init__(self, *args, **kwargs)

        # See the docstring of _get_protocols_status()
        self.protocols_status = {}

        self.routes = []
        self.log = None

    def restart(self):
        """Restart BIRD.

        It runs the "[birdcl/birdcl6] configure" and "restart all" commands.
        """
        if self.reload_config():

            res = self._birdcl("restart all")

            if "restarted" in res:
                return True
            else:
                return False
        else:
            return False

    def reload_config(self):
        """Reload BIRD configuration.

        It runs the "[birdcl/birdcl6] configure" command to reload BIRD's
        configuration.
        """
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        # In BIRD v2, after trying to reload the configuration
        # with 'birdcl configure', I got a weird error:
        #    Reading configuration from /etc/bird/bird.conf
        #    /etc/bird/footer4.local:21:1 Unknown character
        # This trick fixes it.
        self.run_cmd("touch /etc/bird/bird.conf")
        self.run_cmd("touch /etc/bird/footer4.local")
        self.run_cmd("touch /etc/bird/footer6.local")

        res = self._birdcl("configure")

        if "Reconfigured" in res:
            return True
        else:
            return False

    def _get_protocols_status(self, force_update=False):
        """Given the BIRD 'show protocols all' output, build the status map.

        This function fills the self.protocols_status dictionary with entries
        in this format:

        "<BIRD protocol name>": {
            "ip": "<neighbor IP address>",
            "is_up": [True|False]
        }

        This dictionary is used by other functions to understand BGP
        sessions status and to map neighbors IP address to instance name.
        """

        if force_update:
            self.protocols_status = {}

        if self.protocols_status:
            return

        out = self._birdcl("show protocols all")
        lines = out.split("\n")

        last_protoname = None
        last_proto_status = None

        for line in lines:
            if line.strip() and not line.startswith(" "):
                last_protoname = line.split(" ")[0]
                last_proto_status = re.search("\sup\s", line) is not None
                continue

            if line.strip().startswith("Neighbor address: "):
                neighbor_ip = line.strip().split(": ")[1]
                if last_protoname is not None and last_proto_status is not None:
                    self.protocols_status[last_protoname] = {
                        "ip": neighbor_ip,
                        "is_up": last_proto_status
                    }
                    last_protoname = None
                    last_proto_status = None
                else:
                    raise Exception(
                        "{}: neighbor with unknown "
                        "protocol name/status: {}".format(
                            self.name, neighbor_ip
                        )
                    )

        if not self.protocols_status:
            raise Exception(
                "{}: can't build protocols status map".format(self.name)
            )

    def get_bgp_session(self, other_inst_or_ip, force_update=False):
        if isinstance(other_inst_or_ip, BGPSpeakerInstance):
            other_inst_ip = other_inst_or_ip.ip
        else:
            other_inst_ip = other_inst_or_ip
        self._get_protocols_status(force_update=force_update)
        for proto in self.protocols_status:
            if self.protocols_status[proto]["ip"] == other_inst_ip:
                return self.protocols_status[proto]
        return None

    def _get_all_routes(self):

        def add_route(route):
            if not route:
                return

            new_route = Route(**route)
            self.routes.append(new_route)
            route = {}

        self.routes = []

        self._get_protocols_status()

        is_bird2 = "BIRD 2." in self._birdcl("show status")

        if is_bird2:
            # unicast [AS1_1 15:53:50.281] * (100) [AS1i]
            route_beginning_re = (
                "[ ]+unicast "
                "\[(?P<via_name>[^\s]+)"
            )

            # 3.0.11.0/24          unicast [AS3_1 15:53:45.706] * (100) [AS3i]
            prefix_beginning_re = "^(?P<prefix>[0-9\.\:a-f]+/[0-9]+){}".format(
                route_beginning_re
            )
        else:
            # ' via 192.0.2.11 on eth0 [AS1_1 09:57:45] * (100) [AS101i]'
            #       ----------          -----
            route_beginning_re = (
                "[ ]+via "
                "(?P<via>[0-9\.\:a-f]+)"
                "[^\[]+"
                "\[(?P<via_name>[^\s]+)"
            )

            # '101.2.128.0/24     via 192.0.2.11 on eth0 [AS1_1 09:57:45] * (100) [AS101i]'
            #  --------------         ----------          -----
            prefix_beginning_re = "^(?P<prefix>[0-9\.\:a-f]+/[0-9]+){}".format(
                route_beginning_re
            )

        route_beginning_patt = re.compile(route_beginning_re)
        prefix_beginning_patt = re.compile(prefix_beginning_re)

        last_prefix = None
        route = {}
        for option in ["", "filtered"]:
            cmd = "show route all {}".format(option)
            out = self._birdcl(cmd)
            for line in out.split("\n"):
                new_prefix_match = prefix_beginning_patt.search(line)
                new_route_match = route_beginning_patt.search(line)
                if new_prefix_match or new_route_match:
                    # A new route for the last prefix or a block of routes for
                    # a new prefix begins here.
                    add_route(route)
                    route = {}

                    if new_prefix_match:
                        last_prefix = new_prefix_match.group("prefix")
                        via_name = new_prefix_match.group("via_name")
                    else:
                        via_name = new_route_match.group("via_name")

                    route["prefix"] = last_prefix
                    route["via"] = self.protocols_status[via_name]["ip"]
                    route["best"] = (option == "" and " * " in line)
                    route["filtered"] = option == "filtered"
                else:
                    if "BGP.as_path:" in line:
                        raw_as_path = line.split(": ")[1].strip()
                        if "{" in raw_as_path:
                            # Stripping as_set in strings like this:
                            #   BGP.as_path: 222 333 { 333 333 }
                            as_path = raw_as_path[0:raw_as_path.index("{")].strip()
                            as_set = raw_as_path[raw_as_path.index("{") + 1:-1].strip()
                        else:
                            as_path = raw_as_path
                            as_set = None
                        route["as_path"] = as_path
                        route["as_set"] = as_set
                    if "BGP.next_hop:" in line:
                        route["next_hop"] = ""
                        next_hops = line.split(": ")[1].strip().split()
                        for next_hop in next_hops:
                            if next_hop.lower().startswith("fe80:"):
                                continue
                            if route["next_hop"]:
                                route["next_hop"] += ", "
                            route["next_hop"] += next_hop
                    if "BGP.community:" in line:
                        route["std_comms"] = line.split(": ")[1].strip()
                    if "BGP.large_community:" in line:
                        route["lrg_comms"] = line.split(": ")[1].strip()
                    if "BGP.ext_community:" in line:
                        route["ext_comms"] = line.split(": ")[1].strip()
                    if "BGP.local_pref:" in line:
                        route["localpref"] = line.split(": ")[1].strip()
            add_route(route)
            route = {}

    def clear_cached_routes(self):
        self.routes = {}

    def get_routes(self, prefix, include_filtered=False, only_best=False):
        if include_filtered and only_best:
            raise Exception("Can't set both include_filtered and only_best")

        if not self.routes:
            self._get_all_routes()

        routes = []
        for route in self.routes:
            if prefix is None or route.prefix == prefix:
                if route.filtered and not include_filtered:
                    continue
                if not route.best and only_best:
                    continue
                routes.append(route)

        return routes

    def get_protocol_name_by_ip(self, ip):
        self._get_protocols_status()
        for proto in self.protocols_status:
            if self.protocols_status[proto]["ip"] == ip:
                return proto
        raise Exception(
            "{}: can't find protocol name from ip {}".format(self.name, ip)
        )

    def log_contains(self, s):
        if not self.log:
            self.log = self.run_cmd("cat /var/log/bird.log")

        if s in self.log:
            return True
        else:
            return False

    def log_contains_errors(self, allowed_errors=[], list_errors=False):
        out = self.run_cmd("cat /var/log/bird.log")

        errors_found = False
        errors = []
        for line in out.split("\n"):
            if "<ERR>" not in line:
                continue
            if any([msg for msg in allowed_errors if msg in line]):
                continue
            errors_found = True
            errors.append(line)
        if list_errors:
            return errors_found, "\n".join(errors)
        return errors_found

class BIRDInstanceIPv4(BIRDInstance):

    def _get_start_cmd(self):
        return "bird -c /etc/bird/bird.conf -d"

    def _birdcl(self, cmd):
        return self.run_cmd("birdcl {}".format(cmd))

class BIRDInstanceIPv6(BIRDInstance):

    def _get_start_cmd(self):
        return "bird6 -c /etc/bird/bird.conf -d"

    def _birdcl(self, cmd):
        return self.run_cmd("birdcl6 {}".format(cmd))

class BIRD2Instance(BIRDInstance):

    DOCKER_IMAGE = "pierky/bird:2.0.8"

    TAG = "bird2"

    def _get_start_cmd(self):
        return "bird -c /etc/bird/bird.conf -d"

    def _birdcl(self, cmd):
        return self.run_cmd("birdcl {}".format(cmd))



# Copyright (C) 2017-2021 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .docker import DockerInstance
from .instances import InstanceNotRunning, BGPSpeakerInstance


class ExaBGPInstance(DockerInstance, BGPSpeakerInstance):

    MESSAGE_LOGGING_SUPPORT = False

    DOCKER_IMAGE = "pierky/exabgp:4.2.7"

    TAG = "exabgp"

    def restart(self):
        raise NotImplementedError()

    def reload_config(self):
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        # no-op, ExaBPG is only used as a route-server client, no needs
        # to reload its configuration.
        return True

    def get_bgp_session(self, other_inst_or_ip, force_update=False):
        raise NotImplementedError()

    def get_routes(self, prefix, include_filtered=False, only_best=False):
        # no-op, ExaBPG is only used as a route-server client, no needs
        # to dump its routes.
        return []

    def log_contains(self, s):
        raise NotImplementedError()

    def log_contains_errors(self, allowed_errors=[], list_errors=False):
        raise NotImplementedError()

    def _get_start_cmd(self):
        return (
            "exabgp --env /etc/exabgp/exabgp.env /etc/exabgp/exabgp.conf"
        )



# Copyright (C) 2017-2021 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
import difflib
import hashlib
import filecmp
import logging
import os
import six
import sys
import textwrap
import yaml

from ..ask import Ask
from ..irrdb import IRRDBInfo
from ..cached_objects import CachedObject
from ..resources import get_config_dir, get_templates_dir
from ..errors import ConfigError, ARouteServerError, MissingFileError, \
                     ProgramConfigError


class ConfigParserProgram(object):

    DEFAULT_CFG_DIR_USR = "~/arouteserver"
    DEFAULT_CFG_DIR_ETC = "/etc/arouteserver"
    DEFAULT_CFG_DIRS = [DEFAULT_CFG_DIR_ETC,
                        DEFAULT_CFG_DIR_USR]
    DEFAULT_CFG_FILE = "arouteserver.yml"
    DEFAULT_CFG_PATHS = [os.path.join(DEFAULT_CFG_DIR_ETC, DEFAULT_CFG_FILE),
                         os.path.join(DEFAULT_CFG_DIR_USR, DEFAULT_CFG_FILE)]

    DEFAULT = {
        "cfg_dir": None,

        "logging_config_file": "log.ini",

        "cfg_general": "general.yml",
        "cfg_clients": "clients.yml",
        "cfg_bogons": "bogons.yml",

        "templates_dir": "templates",
        "template_name": "main.j2",

        "cache_dir": "cache",
        "cache_expiry": CachedObject.DEFAULT_EXPIRY,

        "bgpq3_path": "bgpq4",
        "bgpq3_host": IRRDBInfo.BGPQ3_DEFAULT_HOST,
        "bgpq3_sources": IRRDBInfo.BGPQ3_DEFAULT_SOURCES,

        "rtt_getter_path": "",

        "threads": 4,

        "check_new_release": True
    }

    PATH_KEYS = ("logging_config_file", "cfg_general", "cfg_clients",
                 "cfg_bogons", "templates_dir", "cache_dir", "rtt_getter_path")

    FINGERPRINTS_FILENAME = "fingerprints.yml"

    def v(self, text, new_line=True):
        if self.verbose:
            sys.stdout.write("{}{}".format(
                text, "\n" if new_line else ""
            ))

    def __init__(self, verbose=True, ask=True):
        self._reset_to_default()
        self.verbose = verbose
        self.ask = ask

    def _reset_to_default(self):
        self.cfg = deepcopy(self.DEFAULT)

    def load(self, path_or_file):
        self._reset_to_default()

        if isinstance(path_or_file, six.string_types):
            path = path_or_file

            if path and not os.path.exists(os.path.expanduser(path)):
                raise MissingFileError(path)

            f = None
        else:
            path = "<file> object"
            f = path_or_file

        try:
            if f:
                cfg_from_file = yaml.safe_load(f.read())
            else:
                with open(os.path.expanduser(path), "r") as f:
                    cfg_from_file = yaml.safe_load(f.read())

            if cfg_from_file:
                for key in cfg_from_file:
                    if key not in ConfigParserProgram.DEFAULT:
                        raise ConfigError(
                            "Unknown statement: {}".format(key)
                        )
                self.cfg.update(cfg_from_file)

        except Exception as e:
            logging.error("An error occurred while reading program "
                          "configuration at {}: {}".format(path, str(e)),
                          exc_info=not isinstance(e, ARouteServerError))
            raise ConfigError()

        if self.cfg["cfg_dir"]:
            self.cfg["cfg_dir"] = os.path.expanduser(self.cfg["cfg_dir"])
        else:
            self.cfg["cfg_dir"] = os.path.dirname(os.path.expanduser(path))

        # relative path -> absolute path
        for cfg_key in self.PATH_KEYS:
            if not self.cfg[cfg_key]:
                continue
            val = os.path.expanduser(self.cfg[cfg_key])
            if not os.path.isabs(val):
                self.cfg[cfg_key] = os.path.join(self.cfg["cfg_dir"], val)
            else:
                self.cfg[cfg_key] = val

    def expanduser(self, cfg_key):
        if cfg_key in self.PATH_KEYS:
            return os.path.expanduser(self.cfg[cfg_key])
        else:
            return self.cfg[cfg_key]

    def parse_cli_args(self, args):
        args_dict = vars(args)
        for option_name in self.cfg:
            if option_name in args_dict and args_dict[option_name]:
                self.cfg[option_name] = args_dict[option_name]

    def get(self, cfg_key):
        if self.cfg[cfg_key]:
            return self.expanduser(cfg_key)
        return None

    def get_dir(self, cfg_key):
        v = self.get(cfg_key)
        if not v:
            raise ProgramConfigError(
                "The value of '{}' is missing.".format(cfg_key)
            )
        if not os.path.exists(v):
            raise ProgramConfigError(
                "The path configured for '{}' ({}) does not exist.".format(
                    cfg_key, v
                )
            )
        if not os.path.isdir(v):
            raise ProgramConfigError(
                "The path configured for '{}' ({}) is not a directory.".format(
                    cfg_key, v
                )
            )
        return v

    def mk_dir(self, d):
        self.v("Creating {}... ".format(d), new_line=False)
        if os.path.exists(d):
            self.v("already exists")
        else:
            try:
                os.makedirs(d)
                self.v("OK")
            except OSError as e:
                raise ARouteServerError(str(e))

    def cp_file(self, s, d):
        if os.path.abspath(s) == os.path.abspath(d):
            return
        # Avoid .swp files
        if s.endswith(".swp"):
            return

        assert os.path.exists(s), "The {} file does not exist.".format(s)

        try:
            with open(s, "r") as src:
                with open(d, "w") as dst:
                    dst.write(src.read())
        except (IOError, OSError) as e:
            raise ARouteServerError(str(e))

    def show_diff(self, s, d):
        with open(s, "r") as f:
            fromlines = f.readlines()
        with open(d, "r") as f:
            tolines = f.readlines()
        diff = difflib.unified_diff(fromlines, tolines,
                                    "currently installed", "new")
        self.v("")
        sys.stdout.writelines(diff)
        self.v("")

    def process_file(self, s, d, fps_status=None, rel_path=None):
        assert os.path.exists(s), "The {} file does not exist.".format(s)

        filename = os.path.basename(s)

        def write_title():
            self.v("- {}... ".format(filename), new_line=False)

        write_title()

        if not os.path.exists(d):
            self.cp_file(s, d)
            self.v("OK (created)")
            return True

        if filecmp.cmp(s, d, shallow=False):
            self.v("skipped (same content)")
            return True

        if fps_status:
            status_descr = ConfigParserProgram.get_fingerprints_status_descr(
                fps_status, rel_path)

            if fps_status["new_file"]:
                self.cp_file(s, d)
                self.v("OK (created)")
                return True

            if fps_status["same_file"]:
                self.v("skipped (same content)")
                return True

            if fps_status["local_unknown"]:
                self.v("WARNING!")
                self.v("")
                self.v(
                    "   "
                    "\n   ".join(textwrap.wrap(status_descr, width=60))
                )
                self.v("")
                write_title()
            else:
                if fps_status["installed_version_mismatch"]:
                    if not fps_status["locally_edited"]:
                        self.cp_file(s, d)
                        self.v("OK (updated)")
                        return True

                if fps_status["locally_edited"]:
                    self.v("WARNING!")
                    self.v("")
                    self.v(
                        "   "
                        "\n   ".join(textwrap.wrap(status_descr, width=60))
                    )
                    self.v("")
                    bak_path = "{}.bak".format(d)

                    if self.ask:
                        ret, yes_no = Ask().ask_yes_no(
                            "Do you want to create "
                            "a backup copy into {}?".format(bak_path),
                            default="yes"
                        )
                    else:
                        ret = True
                        yes_no = "yes"

                    if not ret:
                        return False

                    if yes_no.lower() == "yes":
                        self.cp_file(d, bak_path)
                        self.cp_file(s, d)
                        write_title()
                        self.v("OK (backed up and updated)")
                        return True
                    else:
                        write_title()

        while True:
            if self.ask:
                ret, answer = Ask().ask(
                    "already exists: do you want to overwrite it?",
                    options=["yes", "no", "diff"],
                    default="no"
                )
            else:
                ret = True
                answer = "yes"

            if not ret:
                return False

            if answer == "diff":
                self.show_diff(s, d)
                write_title()
            else:
                write_title()

                if answer != "yes":
                    self.v("skipped")
                    return True

                self.cp_file(s, d)
                self.v("OK")
                return True

    def process_dir(self, s, d, fps_status=None, rel_path=None):
        assert os.path.exists(s) and os.path.isdir(s), \
            "The {} directory does not exist.".format(s)

        self.v("Populating {}...".format(d))

        for filename in os.listdir(s):
            if filename == ConfigParserProgram.FINGERPRINTS_FILENAME:
                continue

            new_fps_status = None
            if fps_status and filename in fps_status:
                new_fps_status = fps_status[filename]
            new_rel_path = None
            if rel_path:
                new_rel_path = os.path.join(rel_path, filename)

            if os.path.isdir(os.path.join(s, filename)):
                self.mk_dir(os.path.join(d, filename))
                if not self.process_dir(
                    os.path.join(s, filename),
                    os.path.join(d, filename),
                    fps_status=new_fps_status,
                    rel_path=new_rel_path
                ):
                    return False
            else:
                if new_fps_status:
                    new_fps_status = new_fps_status["status"]
                if not self.process_file(
                    os.path.join(s, filename),
                    os.path.join(d, filename),
                    fps_status=new_fps_status,
                    rel_path=new_rel_path
                ):
                    return False

        return True

    @staticmethod
    def calculate_fingerprints(d):
        assert os.path.exists(d) and os.path.isdir(d), \
            "The {} directory does not exist.".format(d)

        def iterate_dir(d, dic):
            for filename in os.listdir(d):
                if filename == ConfigParserProgram.FINGERPRINTS_FILENAME:
                    continue
                if filename.endswith(".swp"):
                    continue
                path = os.path.join(d, filename)
                if os.path.isdir(path):
                    dic[filename] = {}
                    iterate_dir(path, dic[filename])
                else:
                    with open(path, "rb") as f:
                        hasher = hashlib.sha512()
                        buf = f.read()
                        hasher.update(buf)
                        dic[filename] = hasher.hexdigest()

        res = {}
        iterate_dir(d, res)
        return res

    @staticmethod
    def load_fingerprints_from_file(path):
        assert os.path.exists(path), "The {} file does not exist.".format(path)

        with open(path, "r") as f:
            return yaml.safe_load(f.read())

    def get_local_fingerprints(self):
        """Calculate fingerprints from local template files."""

        templates_dir = self.get_dir("templates_dir")
        return self.calculate_fingerprints(templates_dir)

    def get_local_distrib_fingerprints(self):
        """Get fingerprints of the locally installed templates.

        Reads the fingerprints from <templates_dir>/<FINGERPRINTS_FILENAME>.

        These fingerprints are those distributed by the program at the time
        of the package installation. A difference between these fingerprints
        and those calculated from the real files means that templates have
        been edited on the local system after the package installation.
        """

        templates_dir = self.get_dir("templates_dir")
        path = os.path.join(templates_dir, self.FINGERPRINTS_FILENAME)
        if os.path.exists(path):
            return self.load_fingerprints_from_file(path)
        return {}

    def get_current_distrib_fingerprints(self):
        """Get fingerprints of the distributed package.

        These fingerprints are those distributed within the current release
        of the package. A difference between these fingerprints and those
        calculated from the real files means that templates have been edited
        on the local system after the package installation or that the current
        release uses different files from those installed on the local system.
        """

        distrib_templates_dir = get_templates_dir()
        path = os.path.join(distrib_templates_dir, self.FINGERPRINTS_FILENAME)
        res = self.load_fingerprints_from_file(path)

        assert res is not None and res != {}, \
            "Empty fingerprints from {}.".format(path)

        return res

    @staticmethod
    def get_fingerprints_status_descr(status, filename):
        if status["new_file"]:
            s = ("{filename} expected but not found on the local "
                 "templates directory")
            return s.format(filename=filename)

        if status["same_file"]:
            s = ("the installed version of {filename} is aligned with "
                 "the one used by the current version of the program")
            return s.format(filename=filename)

        if status["local_unknown"]:
            s = ("the {filename} file is not aligned with the current "
                 "version of the program; since the {fp} file is missing, "
                 "it's not possible to determine if the {filename} file "
                 "has been edited after the installation on the "
                 "local system")
            return s.format(fp=ConfigParserProgram.FINGERPRINTS_FILENAME,
                            filename=filename)

        if status["installed_version_mismatch"]:
            s = ("the installed version of {filename} is not aligned "
                 "with the one used by the current version of the program")
            if status["locally_edited"]:
                s += ("; moreover, it seems that it has been edited "
                        "after the installation on the local system")
            return s.format(filename=filename)

        if status["locally_edited"]:
            s = ("the {filename} file has been edited after the "
                 "installation on the local system")
            return s.format(filename=filename)

        raise NotImplementedError("status: {}".format(str(status)))

    def get_fingerprints_status(self):
        """Build a dict containing the status of a template file.

        new_file is True when there isn't any calculated fingerprint.

        same_file is True when the calculated fingerprint matches the
        fingerprint in the current package.

        local_unknown is True when the local fingerpints.yml file does
        not exists.

        locally_edited is True when the calculated fingerprint does
        not match the fingerprint in the local fingerprints.yml file.

        installed_version_mismatch is True when the fingerprint in the
        local fingerprints.yml file does not match the fingerprint in
        the current package.
        """

        calculated_fps = self.get_local_fingerprints()
        local_distrib_fps = self.get_local_distrib_fingerprints()
        current_distrib_fps = self.get_current_distrib_fingerprints()

        fps_status = {}
        def iterate(curr, local, calc, dst):
            for filename in curr:
                dst[filename] = {}
                if isinstance(curr[filename], dict):
                    iterate(
                        curr[filename],
                        local.get(filename, {}),
                        calc.get(filename, {}),
                        dst[filename]
                    )
                else:
                    dst[filename]["curr"] = curr.get(filename, None)
                    dst[filename]["calc"] = calc.get(filename, None)
                    dst[filename]["local"] = local.get(filename, None)

                    status = {}
                    if dst[filename]["calc"] is None:
                        status["new_file"] = True
                        dst[filename]["status"] = status
                        continue
                    status["new_file"] = False

                    if dst[filename]["calc"] == dst[filename]["curr"]:
                        status["same_file"] = True
                        dst[filename]["status"] = status
                        continue
                    status["same_file"] = False

                    if dst[filename]["local"] is None:
                        status["local_unknown"] = True
                        dst[filename]["status"] = status
                        continue
                    status["local_unknown"] = False

                    status["locally_edited"] = \
                        dst[filename]["local"] != dst[filename]["calc"]
                    status["installed_version_mismatch"] = \
                        dst[filename]["local"] != dst[filename]["curr"]
                    dst[filename]["status"] = status

        iterate(current_distrib_fps, local_distrib_fps, calculated_fps,
                fps_status)
        return fps_status

    def verify_templates(self):
        """Verify if templates are aligned with the current version

        Returns:
            list of errors
        """
        fps_status = self.get_fingerprints_status()

        errors = []

        def iterate(dic, path):
            for filename in dic:
                new_path = os.path.join(path, filename)
                if "status" not in dic[filename]:
                    iterate(dic[filename], new_path)
                else:
                    status = dic[filename]["status"]
                    if not status.get("same_file", False):
                        descr = self.get_fingerprints_status_descr(
                            status, new_path
                        )
                        errors.append(descr)

        iterate(fps_status, "templates")
        return errors

    def setup_templates(self):
        distrib_templates_dir = get_templates_dir()

        # Using .get and not .get_dir because the dest dir may not exist.
        dest_dir = self.get("templates_dir")

        self.v("Installing templates into {}...".format(dest_dir))
        self.v("")

        self.mk_dir(dest_dir)

        fps_status = self.get_fingerprints_status()

        if not self.process_dir(
            distrib_templates_dir, dest_dir, fps_status, "templates"
        ):
            self.v("")
            self.v("Templates installation aborted")
            return False

        self.cp_file(
            os.path.join(distrib_templates_dir, self.FINGERPRINTS_FILENAME),
            os.path.join(dest_dir, self.FINGERPRINTS_FILENAME)
        )
        return True

    def setup(self, destination_directory=None):
        self.v("ARouteServer setup")
        self.v("")

        distrib_config_dir = get_config_dir()

        if destination_directory:
            dest_dir = destination_directory
        else:
            if self.ask:
                res, dest_dir = Ask().ask("Where do you want configuration files and templates "
                                          "to be stored?", default=self.DEFAULT_CFG_DIR_USR)
            else:
                res = True
                dest_dir = self.DEFAULT_CFG_DIR_USR

            if not res:
                self.v("")
                self.v("Setup aborted: no destination directory given")
                return False

        dest_dir = dest_dir.strip()

        if dest_dir not in self.DEFAULT_CFG_DIRS:
            self.v("WARNING: the directory that has been chosen is not one "
                   "of those where the program looks for to find "
                   "its configuration file: use the --cfg command line "
                   "argument to allow the program to find that file.")

        dest_dir = os.path.expanduser(dest_dir)
        program_cfg_file_path = os.path.join(dest_dir, "arouteserver.yml")

        if not destination_directory:
            if self.ask:
                res, yes_or_no = Ask().ask_yes_no(
                    "Do you confirm you want ARouteServer files to be "
                    "stored at {}?".format(dest_dir), default="yes")
            else:
                res = True
                yes_or_no = "yes"

            if not res or yes_or_no.lower() != "yes":
                self.v("")
                self.v("Setup aborted: destination directory not confirmed")
                return False

        self.v("Installing configuration files into {}...".format(dest_dir))
        self.v("")

        self.mk_dir(dest_dir)

        if not self.process_dir(distrib_config_dir, dest_dir):
            self.v("")
            self.v("Setup aborted while populating destination directory")
            return False

        # Load the new configuration, so that the following .setup_templates()
        # can work fine.
        self.load(program_cfg_file_path)

        if not self.setup_templates():
            self.v("")
            self.v("Setup aborted while populating templates directory")
            return False

        # Creating cache directory.
        # Using .get and not .get_dir because the dest dir may not exist.
        cache_dir = self.get("cache_dir")
        self.mk_dir(cache_dir)

        self.v("")
        self.v("Configuration complete!")
        self.v("")
        self.v("- edit the {} file to configure program's options".format(
            program_cfg_file_path))
        self.v("- edit the {} file to set your logging preferences".format(
            self.get("logging_config_file")))
        self.v("- set your route server's options and policies in {}\n"
               "  (edit it manually or use the 'arouteserver configure' "
               "command)".format(
                   self.get("cfg_general")))
        self.v("- configure route server clients in the {} file".format(
            self.get("cfg_clients")))

        return True

program_config = ConfigParserProgram()



# Copyright (C) 2017-2021 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import BaseConfigEnricher
from ..ipaddresses import IPNetwork
from ..errors import BuilderError
from ..ripe_rpki_cache import RIPE_RPKI_ROAs

class RPKIROAsEnricher(BaseConfigEnricher):

    def _use_roas_for_this_origin(self, asn):
        """Return True if ROAs for the given origin ASN must be used

        When ROAs are used only as route objects (so, no BGP Origin
        Validation is configured) only ROAs for authorized origin ASNs
        are returned.
        """

        if self._origin_validation:
            # ROAs are used to perform BGP Origin Validation, so
            # every ROA must be used.
            return True

        if self._roas_as_route_object:
            # If the current ROA is for an origin ASN that is
            # not allowed for any client then skip it.
            if asn in self.origin_asns:
                return True

        return False

    def enrich(self):
        logging.info("Updating RPKI ROAs...")

        filtering = self.builder.cfg_general["filtering"]
        self._roas_as_route_object = \
            filtering["irrdb"]["use_rpki_roas_as_route_objects"]["enabled"]
        self._origin_validation = \
            filtering["rpki_bgp_origin_validation"]["enabled"]

        assert self._roas_as_route_object or self._origin_validation, \
            ("Why here? use_rpki_roas_as_route_objects and "
             "rpki_bgp_origin_validation are both False")

        if self._roas_as_route_object:
            logging.debug("RPKI ROAs will be used as route objects")
        if self._origin_validation:
            logging.debug("RPKI ROAs will be used for origin validation")

        # List of all the origin ASNs.
        if self._roas_as_route_object:
            if self.builder.irrdb_info is None:
                raise BuilderError(
                    "RPKI ROAs can be fetched only after that the "
                    "list of authorized origin ASNs has been built."
                )

            self.origin_asns = set()
            for bundle_id in self.builder.irrdb_info:
                bundle = self.builder.irrdb_info[bundle_id]
                self.origin_asns.update(bundle.asns)

        afis = [4, 6] if self.builder.ip_ver is None else [self.builder.ip_ver]

        rpki_roas_cfg = self.builder.cfg_general["rpki_roas"]
        assert rpki_roas_cfg["source"] == "ripe-rpki-validator-cache", \
            "source is not ripe-rpki-validator-cache"
        urls = rpki_roas_cfg["ripe_rpki_validator_url"]

        ripe_cache = RIPE_RPKI_ROAs(cache_dir=self.builder.cache_dir,
                                    cache_expiry=self.builder.cache_expiry,
                                    ripe_rpki_validator_url=urls)
        ripe_cache.load_data()
        roas = ripe_cache.roas

        allowed_tas = rpki_roas_cfg["allowed_trust_anchors"]

        roas_cnt = {
            "total": 0,
            "invalid_ta": 0,
            "unused": 0,
            "used": {
                "4": 0,
                "6": 0
            }
        }
        for roa in roas["roas"]:
            roas_cnt["total"] += 1

            asn = int(roa["asn"][2:])
            if not self._use_roas_for_this_origin(asn):
                roas_cnt["unused"] += 1
                continue

            ta = roa["ta"]
            if ta not in allowed_tas:
                roas_cnt["invalid_ta"] += 1
                continue

            prefix = roa["prefix"]

            prefix_obj = IPNetwork(prefix)
            if prefix_obj.version not in afis:
                continue

            max_len = int(roa["maxLength"])

            roa_payload = {"prefix": prefix, "length": prefix_obj.prefixlen,
                           "max_len": max_len, "asn": asn}

            prefix_len = str(prefix_obj.prefixlen)
            if prefix_len not in self.builder.rpki_roas:
                self.builder.rpki_roas[prefix_len] = [roa_payload]
            else:
                self.builder.rpki_roas[prefix_len].append(roa_payload)

            roas_cnt["used"][str(prefix_obj.version)] += 1

        stats = "RPKI ROAs: "
        stats += "{} total".format(roas_cnt["total"])
        if roas_cnt["invalid_ta"] > 0:
            stats += ", {} from not allowed TAs".format(roas_cnt["invalid_ta"])
        if roas_cnt["unused"] > 0:
            stats += ", {} unused".format(roas_cnt["unused"])
        for afi in afis:
            stats += ", {} used for IPv{}".format(
                roas_cnt["used"][str(afi)], afi
            )
        logging.info(stats)

        del roas



# Copyright (C) 2017-2021 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
from collections import OrderedDict
import os
from packaging import version
import sys
import textwrap
import yaml

from .base import ARouteServerCommand

from ..ask import Ask
from ..builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from ..config.program import program_config
from ..errors import ARouteServerError
from ..ipaddresses import IPNetwork

class ConfigureCommand(ARouteServerCommand):

    COMMAND_NAME = "configure"
    COMMAND_HELP = ("Initialize a working 'general.yml' file "
                    "with suggested route server's policies and "
                    "settings. "
                    "Examples can be found at "
                    "https://github.com/pierky/arouteserver/tree/"
                    "master/examples/auto-config")

    NEEDS_CONFIG = True

    def __init__(self, *args, **kwargs):
        ARouteServerCommand.__init__(self, *args, **kwargs)

        self.answers = {}
        self.notes = []
        self.preset_answers = {}

    @classmethod
    def add_arguments(cls, parser):
        super(ConfigureCommand, cls).add_arguments(parser)

        parser.add_argument(
            "-o", "--output",
            help="Output file. "
                 "By default, the general.yml file referenced by the "
                 "program configuration will be used. "
                 "Use '-' for stdout.",
            dest="output_file")

        parser.add_argument(
            "--preset-answer",
            nargs="+",
            help=argparse.SUPPRESS,
            dest="preset_answer"
        )

    def wr_text(self, s, title=None, dont_wrap=False):
        sys.stdout.write("\n")
        if title:
            sys.stdout.write(title + "\n")
            sys.stdout.write("=" * len(title) + "\n")
            sys.stdout.write("\n")
        if not s:
            return
        for line in s.split("\n"):
            if dont_wrap:
                sys.stdout.write(line)
            else:
                sys.stdout.write("\n".join(textwrap.wrap(line, width=80)))
            sys.stdout.write("\n")
        sys.stdout.write("\n")

    def add_answer(self, answer_id, ask_func, *args, **kwargs):
        if self.preset_answers:
            self.ask.next_answer = self.preset_answers[answer_id]

        answer_ok = False
        while not answer_ok:
            try:
                answer_ok, answer = ask_func(*args, raise_exc=True, **kwargs)
            except KeyboardInterrupt:
                raise ARouteServerError("Aborted!")
            if answer_ok:
                break
            else:
                if self.preset_answers:
                    raise ARouteServerError("Pre-set answer is invalid!")
        self.answers[answer_id] = answer

    def ask_list_of_ip_prefixes(self, text, raise_exc=False):
        answer_given, v = self.ask.ask(text, None, None, raise_exc)
        if not answer_given:
            return False, None
        res = []
        for s in v.split(","):
            try:
                prefix = IPNetwork(s.strip())
                res.append("{}/{}".format(prefix.ip, prefix.prefixlen))
            except:
                print("Invalid input: a list of IP prefixes is expected. "
                      "'{}' is invalid.".format(s))
                return False, None
        return True, res

    #def ask_16bit_int(self, text, raise_exc=False):
    #    answer_given, v = self.ask.ask_int(text, None, None, raise_exc)
    #    if not answer_given:
    #        return False, None
    #    if v > 65535:
    #        print("Invalid input: a 16-bit integer is expected.")
    #        return False, None
    #    return True, v

    def collect_answers(self):
        if self.answers:
            return

        self.ask = Ask()

        self.wr_text(
            "Depending on the BGP daemon used for the route server "
            "some features may not be available.\n"
            "\n"
            "Details here:\n"
            "https://arouteserver.readthedocs.io/en/latest/"
            "CONFIG.html#caveats-and-limitations",
            title="BGP daemon"
        )
        self.add_answer("daemon", self.ask.ask,
            "Which BGP daemon will be used?",
            options=["bird", "openbgpd"]
        )

        if self.answers["daemon"] == "openbgpd":
            self.add_answer("version", self.ask.ask,
                "Which version?",
                options=OpenBGPDConfigBuilder.AVAILABLE_VERSION,
                default=OpenBGPDConfigBuilder.AVAILABLE_VERSION[-1]
            )
        else:
            self.add_answer("version", self.ask.ask,
                "Which version?",
                options=BIRDConfigBuilder.AVAILABLE_VERSION,
                default=BIRDConfigBuilder.DEFAULT_VERSION
            )

        self.wr_text(
            None, title="Router server's ASN"
        )
        self.add_answer("asn", self.ask.ask_int,
            "What's the ASN of the route server?"
        )

        if self.answers["asn"] > 65535:
            self.wr_text(
                "Since the ASN used for the route server is "
                "a 32-bit value, features and services offered "
                "to clients using BGP communities "
                "will be configured to use a placeholder 16 bit "
                "ASN: 65534",
                #"can be "
                #"automatically configured only if a 16-bit "
                #"placeholder ASN is also given "
                #"(with the exception of BGP Large Communities).",
                title="Place-holder 16-bit ASN for BGP communities"
            )
            #self.add_answer("comms_asn", self.ask_16bit_int,
            #    "16-bit ASN used for BGP communities"
            #)
            self.answers["comms_asn"] = 65534

        self.wr_text(
            None, title="Route server's BGP router-id"
        )
        self.add_answer("router_id", self.ask.ask_ipv4_addr,
            "Please enter the route server BGP router-id:"
        )

        self.wr_text(
            "A list of local IPv4/IPv6 networks must be provided "
            "here: routes announced by route server clients for "
            "these prefixes will be filtered out.",
            title="List of local networks"
        )
        self.add_answer("black_list", self.ask_list_of_ip_prefixes,
            "Please enter a comma-separated list of local networks:"
        )

        return self.answers

    def process_answers(self):

        def add_comm(name, std=None, lrg=None):
            if self.answers["asn"] > 65535:
                rs_as = self.answers["comms_asn"]
            else:
                rs_as = "rs_as"

            if name not in cfg["communities"]:
                cfg["communities"][name] = OrderedDict()

            if self.answers["daemon"] == "openbgpd":
                if version.parse(self.answers["version"]) > version.parse("6.0"):
                    communities = (
                        ("std", std),
                        ("lrg", lrg)
                    )
                else:
                    communities = (
                        ("std", std),
                    )
            else:
                communities = (
                    ("std", std),
                    ("ext", "rt:{}".format(std)),
                    ("lrg", lrg)
                )

            for comm_type, comm_val in communities:
                if comm_val:
                    cfg["communities"][name][comm_type] = \
                        comm_val.format(
                            rs_as=rs_as
                        )

        res = OrderedDict()
        res["cfg"] = OrderedDict()
        cfg = res["cfg"]

        cfg["rs_as"] = self.answers["asn"]

        cfg["router_id"] = self.answers["router_id"]

        if self.answers["daemon"] == "openbgpd":
            self.notes.append(
                "For OpenBGPD, path-hiding mitigation techniques are "
                "not implemented."
            )
            cfg["path_hiding"] = False

        cfg["filtering"] = OrderedDict()
        filtering = cfg["filtering"]

        filtering["next_hop"] = {"policy": "strict"}

        filtering["ipv4_pref_len"] = {"min": 8, "max": 24}
        filtering["ipv6_pref_len"] = {"min": 12, "max": 48}
        self.notes.append(
            "Accepted prefix lengths are 8-24 for IPv6 and 12-48 for IPv6."
        )

        filtering["global_black_list_pref"] = []
        for net in [IPNetwork(_) for _ in self.answers["black_list"]]:
            entry = OrderedDict()
            entry["prefix"] = str(net.ip)
            entry["length"] = net.prefixlen
            filtering["global_black_list_pref"].append(
                entry
            )

        filtering["max_as_path_len"] = 32

        filtering["reject_invalid_as_in_as_path"] = True

        filtering["transit_free"] = {
            "action": "reject",
            "asns": [174, 209, 286, 701, 1239, 1299, 2828, 2914,
                     3257, 3320, 3356, 3549, 5511, 6453, 6461,
                     6762, 6830, 7018, 12956]
        }
        filtering["never_via_route_servers"] = {
            "peering_db": True
        }
        self.notes.append(
            "Routes with 'transit-free networks' or "
            "'never via route-server' (PeeringDB) ASNs in the middle of "
            "AS_PATH are rejected."
        )

        filtering["irrdb"] = OrderedDict()
        irrdb = filtering["irrdb"]
        irrdb["enforce_origin_in_as_set"] = True
        irrdb["enforce_prefix_in_as_set"] = True
        irrdb["allow_longer_prefixes"] = True
        self.notes.append(
            "IRR-based filters are enabled; prefixes that are more specific "
            "of those registered are accepted."
        )
        irrdb["tag_as_set"] = True
        irrdb["peering_db"] = True
        self.notes.append(
            "PeeringDB is used to fetch AS-SETs for those clients that are "
            "not explicitly configured."
        )
        irrdb["use_rpki_roas_as_route_objects"] = {
            "enabled": True
        }
        self.notes.append(
            "RPKI ROAs are used as if they were route objects to further "
            "enrich IRR data."
        )
        irrdb["use_arin_bulk_whois_data"] = {
            "enabled": True
        }
        irrdb["use_registrobr_bulk_whois_data"] = {
            "enabled": True
        }
        self.notes.append(
            "ARIN Whois database dump is fetched from NLNOG to further "
            "enrich IRR data."
        )
        self.notes.append(
            "NIC.BR Whois database dump is fetched from Registro.br to further "
            "enrich IRR data."
        )

        filtering["rpki_bgp_origin_validation"] = OrderedDict()
        filtering["rpki_bgp_origin_validation"]["enabled"] = True
        filtering["rpki_bgp_origin_validation"]["reject_invalid"] = True
        self.notes.append(
            "RPKI BGP Origin Validation is enabled. INVALID routes are "
            "rejected."
        )

        filtering["max_prefix"] = {
            "action": "shutdown",
            "peering_db": {
                "enabled": True
            }
        }
        self.notes.append(
            "PeeringDB is used to fetch networks prefix count."
        )

        if self.answers["daemon"] == "bird":
            filtering["reject_policy"] = {
                "policy": "tag_and_reject"
            }

        if (
            self.answers["daemon"] == "bird" and \
            version.parse(self.answers["version"]) >= version.parse("2.0") and \
            version.parse(self.answers["version"]) <= version.parse("2.0.7")
        ):
            # Why? See comments in v1.0.1 about the 'receive limit' bug in 2.0.7.
            filtering["max_prefix"]["count_rejected_routes"] = False
            self.notes.append(
                "Rejected routes are not counted towards max-prefix limit."
            )

        cfg["graceful_shutdown"] = {"enabled": False}
        if self.answers["daemon"] == "bird":
            cfg["graceful_shutdown"] = {"enabled": True}
        if self.answers["daemon"] == "openbgpd" and \
            version.parse(self.answers["version"]) >= version.parse("6.2"):
            cfg["graceful_shutdown"] = {"enabled": True}
        if cfg["graceful_shutdown"]["enabled"]:
            self.notes.append(
                "Routes tagged with the GRACEFUL_SHUTDOWN well-known "
                "community (65535:0) are processed accordingly to "
                "draft-ietf-grow-bgp-gshut."
            )

        cfg["rfc1997_wellknown_communities"] = {"policy": "pass"}

        cfg["communities"] = OrderedDict()

        if self.answers["daemon"] == "bird":
            add_comm("reject_cause",
                     "65520:dyn_val", "rs_as:65520:dyn_val")

        add_comm("prefix_present_in_as_set",
                 "64512:11", "rs_as:64512:11")
        add_comm("prefix_not_present_in_as_set",
                 "64512:10", "rs_as:64512:10")
        add_comm("origin_present_in_as_set",
                 "64512:21", "rs_as:64512:21")
        add_comm("origin_not_present_in_as_set",
                 "64512:20", "rs_as:64512:20")
        add_comm("prefix_validated_via_rpki_roas",
                 "64512:31", "rs_as:64512:31")
        add_comm("route_validated_via_white_list",
                 "64512:41", "rs_as:64512:41")

        add_comm("do_not_announce_to_any",
                 "0:{rs_as}", "rs_as:0:0")
        add_comm("do_not_announce_to_peer",
                 "0:peer_as", "rs_as:0:peer_as")
        add_comm("announce_to_peer",
                 "{rs_as}:peer_as", "rs_as:1:peer_as")

        add_comm("prepend_once_to_any",
                 "65501:{rs_as}", "rs_as:101:0")
        add_comm("prepend_twice_to_any",
                 "65502:{rs_as}", "rs_as:102:0")
        add_comm("prepend_thrice_to_any",
                 "65503:{rs_as}", "rs_as:103:0")

        add_comm("prepend_once_to_peer",
                 "65511:peer_as", "rs_as:101:peer_as")
        add_comm("prepend_twice_to_peer",
                 "65512:peer_as", "rs_as:102:peer_as")
        add_comm("prepend_thrice_to_peer",
                 "65513:peer_as", "rs_as:103:peer_as")

        add_comm("add_noexport_to_peer",
                 "65281:peer_as", "rs_as:65281:peer_as")
        add_comm("add_noadvertise_to_peer",
                 "65282:peer_as", "rs_as:65282:peer_as")

        return res

    def configure_yml(self):

        def represent_ordereddict(dumper, data):
            value = []

            for item_key, item_value in data.items():
                node_key = dumper.represent_data(item_key)
                node_value = dumper.represent_data(item_value)

                value.append((node_key, node_value))

            return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

        self.collect_answers()
        cfg = self.process_answers()

        yaml.add_representer(OrderedDict, represent_ordereddict)
        return yaml.dump(cfg, default_flow_style=False)

    def configure_dict(self):
        self.collect_answers()
        return self.process_answers()

    def run(self):
        self.preset_answers = {}
        if self.args.preset_answer:
            for key_val in self.args.preset_answer:
                if not key_val.strip():
                    continue
                preset_key = key_val.split("=")[0]
                preset_val = key_val.split("=")[1]
                self.preset_answers[preset_key] = preset_val

        dest = self.args.output_file
        if not dest:
            dest = program_config.get("cfg_general")

        if dest != "-":
            if os.path.exists(dest):
                _, yes_no = Ask().ask_yes_no(
                    "The file {} already exists; "
                    "it will be overwritten: "
                    "proceed anyway?".format(dest),
                    default="no"
                )
                if yes_no.lower() != "yes":
                    print("Aborted!")
                    return

        yml = self.configure_yml()

        msg = "\n\n"
        msg += "Route server policy definition file generated successfully!\n"
        msg += "===========================================================\n"
        msg += "\n"
        msg += "The content of the general configuration file "
        if dest == "-":
            msg += "follows.\n"
        else:
            msg += "will now be written to {}\n".format(dest)
        msg += "\n"
        msg += "Some notes:\n"
        msg += "\n"
        msg += " - "
        msg += "\n - ".join(self.notes)

        if dest == "-":
            msg = "# " + "\n# ".join([line for line in msg.split("\n")])
            self.wr_text(msg, dont_wrap=True)
            sys.stdout.write(yml)
        else:
            self.wr_text(msg)
            try:
                with open(dest, "w") as f:
                    f.write(yml)
            except Exception as e:
                raise ARouteServerError(
                    "An error occurred while writing the output "
                    "configuration to {} - {}".format(
                        dest, str(e)
                    )
                )

        return True
